/*
 * Created on 12-Nov-06
 *
 * To change the template for this generated file go to
 * Window - Preferences - Java - Code Generation - Code and Comments
 */
package abc.tm.weaving.weaver.tmanalysis.query;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import polyglot.util.ErrorInfo;
import soot.Local;
import soot.PointsToSet;
import soot.PrimType;
import soot.Scene;
import soot.SootMethod;
import soot.jimple.toolkits.pointer.FullObjectSet;
import abc.tm.weaving.weaver.tmanalysis.MatchingTMSymbolTag;
import abc.tm.weaving.weaver.tmanalysis.ds.PaddlePointsToSetCompatibilityWrapper;
import abc.weaving.aspectinfo.Var;
import abc.weaving.residues.WeavingVar;

/**
 * Shadow
 *
 * @author Eric Bodden
 */
public class Shadow {
	
	protected Map varToPointsToSet;		

	protected String uniqueShadowId;

	protected boolean hasEmptyMapping;
	
	/**
	 * @param variableMapping
	 */
	public Shadow(String uniqueShadowId, Map variableMapping, SootMethod container) {
		assert uniqueShadowId!=null && variableMapping!=null && container!=null;
		this.uniqueShadowId = uniqueShadowId;
		this.hasEmptyMapping = false;
		importVariableMapping(variableMapping, container);
	}
	
	public static Set allShadowsForTag(MatchingTMSymbolTag tag, SootMethod container) {
		Set result = new HashSet();
		for (Iterator symbolIdIter = tag.getMatchingSymbolIDs().iterator(); symbolIdIter.hasNext();) {
			String symbolId = (String) symbolIdIter.next();
			
			Map mapping = tag.getVariableMappingForSymbol(symbolId);
			String uniqueShadowId = tag.getMatchingUniqueShadowID(symbolId);
			
			result.add(new Shadow(uniqueShadowId,mapping,container));
		}
		return result;
	}
	
	public void importVariableMapping(Map mapping, SootMethod container)  {
		if(varToPointsToSet!=null) {
			throw new IllegalStateException("mapping already set");
		}
		varToPointsToSet = new HashMap();
		//for each WeavingVar, get its local l and for that l get and store
		//the points-to set
		for (Iterator iter = mapping.entrySet().iterator(); iter.hasNext();) {
			Entry entry = (Entry) iter.next();
			Collection weavingVars = (Collection) entry.getValue();

			assert weavingVars.size()<=1;

			if(weavingVars.size()>0) {
				WeavingVar wv = (WeavingVar) weavingVars.iterator().next();
				Local l = wv.get();
				
				PointsToSet pts;
				if(l.getType() instanceof PrimType) {
					//if the type of the variable is a primitive type, then we assume it could "point to anything", i.e. could have any value
					pts = FullObjectSet.v();
				} else {
					//if l is null this probably means that the WeavingVar was never woven
					assert l!=null;
					PointsToSet paddlePts = (PointsToSet) Scene.v().getPointsToAnalysis().reachingObjects(l);
					if(paddlePts.isEmpty()) {
						hasEmptyMapping = true;
						abc.main.Main.v().error_queue.enqueue(ErrorInfo.WARNING, "Empty points-to set for variable "+l+" in "+container);
						/*   POSSIBLE CAUSES FOR EMPTY VARIABLE MAPPINGS:
						 *   1.) a shadow is created for an invoke sattement (or similar) of the form o.foo() where o has type NullType, i.e. is certainly null.
						 *   2.) if object-sensitivity is enabled:
						 *       if dynamic type checks for advice aplication can be ruled out statically; 
						 *        example: point cut is call(* Set.add(..)) && args(Collection)
						 *                 shadow is s.add("someString")
						 *                 in this case, paddle sees automatically that the instanceof check in the bytecode
						 *                 can never succeed; hence the statement is rendered unreachable and the points-to set becomes empty
						 */
					}							
					pts = new PaddlePointsToSetCompatibilityWrapper(paddlePts);
				}
				varToPointsToSet.put(((Var)entry.getKey()).getName(),pts);
			}
		}
	}
	
	/**
	 * @return the varToPointsToSet
	 */
	public Map getVariableMapping() {
		return Collections.unmodifiableMap(new HashMap(varToPointsToSet));
	}

	
	public Set getBoundVariables() {
		return Collections.unmodifiableSet(new HashSet(varToPointsToSet.keySet()));
	}
	;
	public boolean hasVariableMapping() {
		return varToPointsToSet!=null;
	}

	
	/**
	 * TODO comment
	 * @param v
	 * @return
	 */
	public PointsToSet getPointsToSet(String v) {
		return (PointsToSet) varToPointsToSet.get(v);
	}
	
	/**
	 * @return the uniqueShadowId
	 */
	public String getUniqueShadowId() {
		return uniqueShadowId;
	}
	
	/**
	 * For a given set of {@link Shadow}s, returns the set of their
	 * unique shadow IDs.
	 * @param shadows a set of {@link Shadow}s
	 * @return the set of their shadow IDs
	 * @see Naming#uniqueShadowID(String, int)
	 * @see Naming#uniqueShadowID(String, String, int)
	 */
	public static Set uniqueShadowIDsOf(Set shadows) {
		Set ids = new HashSet();
		for (Iterator shadowIter = shadows.iterator(); shadowIter.hasNext();) {
			Shadow shadow = (Shadow) shadowIter.next();
			ids.add(shadow.getUniqueShadowId());
		}
		return Collections.unmodifiableSet(ids);
	}

	/**
	 * {@inheritDoc}
	 */
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((uniqueShadowId == null) ? 0 : uniqueShadowId.hashCode());
		result = prime
				* result
				+ ((varToPointsToSet == null) ? 0 : varToPointsToSet.hashCode());
		return result;
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final Shadow other = (Shadow) obj;
		if (uniqueShadowId == null) {
			if (other.uniqueShadowId != null)
				return false;
		} else if (!uniqueShadowId.equals(other.uniqueShadowId))
			return false;
		if (varToPointsToSet == null) {
			if (other.varToPointsToSet != null)
				return false;
		} else if (!varToPointsToSet.equals(other.varToPointsToSet))
			return false;
		return true;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public String toString() {
		return uniqueShadowId + "(" + varToPointsToSet + ")";
	}

	/**
	 * @return the hasEmptyMapping
	 */
	public boolean hasEmptyMapping() {
		return hasEmptyMapping;
	}
	
	

}
