/*
 * Created on 13-Nov-06
 *
 * To change the template for this generated file go to
 * Window - Preferences - Java - Code Generation - Code and Comments
 */
package abc.tm.weaving.weaver.tmanalysis.stages;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import soot.Timers;
import abc.main.AbcTimer;
import abc.main.Main;
import abc.tm.weaving.aspectinfo.TMGlobalAspectInfo;
import abc.tm.weaving.matching.SMEdge;
import abc.tm.weaving.matching.TMStateMachine;
import abc.tm.weaving.weaver.tmanalysis.ComponentStateMachineConstructor;
import abc.tm.weaving.weaver.tmanalysis.Timer;
import abc.tm.weaving.weaver.tmanalysis.UGStateMachine;
import abc.tm.weaving.weaver.tmanalysis.WorklistBasedTMFlowAnalysis;
import abc.tm.weaving.weaver.tmanalysis.VariableSMEdgeFactory.SMVariableEdge;
import abc.tm.weaving.weaver.tmanalysis.query.Shadow;
import abc.tm.weaving.weaver.tmanalysis.query.ShadowGroup;
import abc.tm.weaving.weaver.tmanalysis.query.ShadowRegistry;
import abc.tm.weaving.weaver.tmanalysis.query.StateMachineFolder;

/**
 * Flow-sensitive analysis stage. This should usually be applied after {@link FlowInsensitiveAnalysis} for improved efficiency, but this is not a strict requirement.
 * This stage builds an interprocedural abstraction and the performs a fixed-point iteration over it.
 *
 * @author Eric Bodden
 */
public class FlowSensitiveAnalysis extends AbstractAnalysisStage {

	/** shortcut to the global aspect info */
	protected static TMGlobalAspectInfo gai = (TMGlobalAspectInfo) Main.v().getAbcExtension().getGlobalAspectInfo();

	/** the state machine representing the complete program */
	protected TMStateMachine completeStateMachine;

	/** a timer to time the duration of the construction of the interprocedural abstraction */
	protected Timer abstrTimer = new Timer("interpr-abstraction");
	
	/** a timer to time the duration of the fixed point analysis */
	protected Timer fpTimer = new Timer("fixed-point-iteration");
	
	protected transient boolean flowSensitiveAnalysisDone = false;
	
	protected int numLabelShadowsOnly = 0;

	/**
	 * {@inheritDoc}
	 */
	protected void doAnalysis() {

		buildInterproceduralAbstraction();
		
		filterEdgesReferringToSharedVariables();

		//fixedPointIteration();
	}
	
	/**
	 * Retains edges not affecting the main thread, just to be sound.
	 */
	protected void filterEdgesReferringToSharedVariables() {
		Map shadowIdToEdges = new HashMap();
		
		//build mapping from shadow IDs to all edges associated with that shadow id
		for (Iterator edgeIter = completeStateMachine.getEdgeIterator(); edgeIter.hasNext();) {
			SMEdge edge = (SMEdge) edgeIter.next();
			if(edge instanceof SMVariableEdge && edge.getLabel()!=UGStateMachine.EPSILON) {
				SMVariableEdge variableEdge = (SMVariableEdge) edge;
				String shadowId = variableEdge.getQualifiedShadowId();
				Set edges = (Set) shadowIdToEdges.get(shadowId);
				if(edges==null) {
					edges = new HashSet();
					shadowIdToEdges.put(shadowId, edges);
				}
				edges.add(edge);
			}
		}
		
		Set consistentShadowGroups = FlowInsensitiveAnalysis.v().getAllConsistentShadowGroups();
		Set shadowsToRetain = new HashSet();
		Set singleThreadGroups = new HashSet();
		
		int singleThreadedGroups = 0;
		int multiThreadedGroups = 0;
		
		//retain all shadows of group with shared variables
		for (Iterator groupIter = consistentShadowGroups.iterator(); groupIter.hasNext();) {
			ShadowGroup group = (ShadowGroup) groupIter.next();
			
			if(!group.getTraceMatch().isPerThread() && group.hasSharedVariables()) {
				shadowsToRetain.addAll(group.getAllShadows());
				multiThreadedGroups++;
			} else {
				singleThreadGroups.add(group);
				singleThreadedGroups++;
			}
		}
		logToStatistics("shadow-groups", consistentShadowGroups.size());
		logToStatistics("single-threaded-shadow-groups", singleThreadedGroups);
		logToStatistics("multi-threaded-shadow-groups", multiThreadedGroups);
		
		Set idsOfShadowsToRetain = Shadow.uniqueShadowIDsOf(shadowsToRetain);
		retainAll(idsOfShadowsToRetain);	
		
		//remove all edges from the complete state machine which point to edges with shared variables
		for (Iterator idIter = idsOfShadowsToRetain.iterator(); idIter.hasNext();) {
			String shadowId = (String) idIter.next();
			Collection edges = (Collection) shadowIdToEdges.get(shadowId);
			assert edges!=null; //for each shadow there should be at least one edge
			for (Iterator edgeIter = edges.iterator(); edgeIter.hasNext();) {
				SMEdge edge = (SMEdge) edgeIter.next();
				completeStateMachine.replaceEdge(edge, new SMEdge(edge.getSource(),edge.getTarget(),UGStateMachine.EPSILON));
			}
		}		
		completeStateMachine.cleanup();
		completeStateMachine.minimizeIfSmaller();

		Set toRetain = new HashSet();
		//build component state machines
		for (Iterator groupIter = singleThreadGroups.iterator(); groupIter.hasNext();) {
			ShadowGroup group = (ShadowGroup) groupIter.next();
			fpTimer.startOrResume();
			Set activeShadows = new HashSet(perTracematchFixedPointIteration(group));
			System.err.println("label "+group.getLabelShadows().size());
			System.err.println("skip  "+group.getSkipShadows().size());
//			activeShadows.addAll(Shadow.uniqueShadowIDsOf(group.getSkipShadows()));
			fpTimer.stop();
			toRetain.addAll(activeShadows);
		}
		
		logToStatistics("cum-fp-timer", fpTimer);
		
		logToStatistics("fp-iterations-label-shadows-only", numLabelShadowsOnly);

		retainAll(toRetain);
	}

	/**
	 * Builds an interprocedural representation of the whole program by combining {@link UGStateMachine}s based on the
	 * abstracted call graph.
	 * @see CallGraphAbstraction
	 */
	protected void buildInterproceduralAbstraction() {
		
		abstrTimer.startOrResume();
		completeStateMachine = StateMachineFolder.v().getCompleteProgramStateMachine(CallGraphAbstraction.v().abstractedCallGraph());
		abstrTimer.stop();
		logToStatistics("interproc-abstr-timer", abstrTimer);
		logToStatistics("size-program-graph", completeStateMachine.size());
		
		//UGStateMachine.statistics();
		
		AbcTimer.mark("TMAnalysis: fold interprocedurally");
	}

//	/**
//	 * Performs a fixed point iteration over the whole program representation in order to find traces of shadows that might
//	 * drive a tracematch into a final state.
//	 */
//	protected void fixedPointIteration() {
////		final long TIMEOUT = 60000;
////		
////		final Object lock = new Object();
//		
//		final Set activeShadows = Collections.synchronizedSet(new HashSet());
//		
////		Thread fpThread = new Thread() {
////			public void run() {
//				for (Iterator tmIter = gai.getTraceMatches().iterator(); tmIter.hasNext();) {
//					TraceMatch tm = (TraceMatch) tmIter.next();
//
//					fpTimer.startOrResume();
//					Set thisTmActiveShadows = perTracematchFixedPointIteration(tm, completeStateMachine);
//					fpTimer.stop();
//					activeShadows.addAll(thisTmActiveShadows);enko auch Beresowski in dessen Londoner Büro besucht haben. Dort fand Scotland Yard später ra
//				}
////				synchronized (lock) {
////					flowSensitiveAnalysisDone = true;
////				}
////			}
////		};
////		
////		fpThread.start();
////		
////		synchronized (fpThread) {
////			try {
////				fpThread.join(TIMEOUT);
////			} catch (InterruptedException e) {
////				e.printStackTrace();
////			}
////		}		
////		
//		logToStatistics("cum-fp-timer", fpTimer);
//		
////		synchronized (lock) {
////			//if the analysis finished in time, retain only the active shadows, else all
////			if(flowSensitiveAnalysisDone) {
//				retainAll(activeShadows);
////				logToStatistics("flow-sensitive-analysis-aborted", "false");
////			} else {
////				retainAll(ShadowRegistry.v().enabledShadows());
////				System.err.println("flow-sensitive analysis timed out after "+TIMEOUT+"ms");
////				logToStatistics("flow-sensitive-analysis-aborted", "true");
////			}
////		}
//	}
	
	/**
	 * @param group
	 * @param fullIteration
	 * @param defaultOrder
	 * @return
	 */
	protected Set perTracematchFixedPointIteration(ShadowGroup group) {
		
		//see if all required shadows are still present in the state machine;
		//(some could have been removed because they were shared by multiple threads)
		Set labelShadowsToFind = new HashSet(group.getLabelShadows());
		
		for (Iterator edgeIter = completeStateMachine.getEdgeIterator(); edgeIter.hasNext();) {
			SMEdge edge = (SMEdge) edgeIter.next();
			if(edge instanceof SMVariableEdge) {
				SMVariableEdge variableEdge = (SMVariableEdge) edge;
				if(variableEdge.hasShadow()) {
					labelShadowsToFind.remove(variableEdge.getShadow());
					if(labelShadowsToFind.isEmpty()) {
						break;
					}
				}
			}			
		}
		
		boolean allShadowsPresent = labelShadowsToFind.isEmpty();
		
		//if there are shadows missing for a complete match, return the empty set
		if(!allShadowsPresent) {
			return Collections.EMPTY_SET;
		}

		//if there are more than twice as many skip shadows than label shadows, consider label shadows only
		boolean onlyConsiderLabelShadows = group.getSkipShadows().size() > group.getLabelShadows().size() * 2;
		System.err.println("-------------------------------------------------------------------");
		if(onlyConsiderLabelShadows) {
			System.err.println("performing fixed point iteration considering label shadows only");
			numLabelShadowsOnly ++;
		} else {
			System.err.println("performing full fixed point iteration");
		}
		System.err.println("num-label-shadows="+group.getLabelShadows().size());
		System.err.println("num-skip-shadows="+group.getSkipShadows().size());

		int startcount = Timers.v().totalFlowComputations;
		Set result = new WorklistBasedTMFlowAnalysis(group, completeStateMachine, onlyConsiderLabelShadows).getActiveShadows();
		int fpCompCount = Timers.v().totalFlowComputations - startcount;
		
		System.err.println(group.getTraceMatch().getName()+"-flow-computation-count="+fpCompCount);
		System.err.println(group.getTraceMatch().getName()+"-active-shadow-count="+result.size());
		
		return result;
	}
	
	protected static FlowSensitiveAnalysis instance;

	private FlowSensitiveAnalysis() {}
	
	public static FlowSensitiveAnalysis v() {
		if(instance==null) {
			instance = new FlowSensitiveAnalysis();
		}
		return instance;		
	}

}
